import os
import re
import yaml
import builds
from pathlib import Path
from argparse import ArgumentParser

from .remote import Remote
from .printing import print


class UInstall:

    def __init__(self,
                 preset: Path = None,
                 host: str = None,
                 user: str = None,
                 password: str = None,
                 public_key: str = None,
                 sudo: bool = False,
                 debug: bool = False):

        self.preset_data = {}
        if preset:
            with print('🟪[Loading preset]', close='\n'):
                self.load_preset(preset)

        self.debug = debug

        if self.preset_data.get('remote', None):
            remote_preset = self.preset_data['remote']
            host = host or remote_preset.get('host', None)
            user = user or remote_preset.get('user', None)
            public_key = public_key or remote_preset.get('public_key', None)
            sudo = sudo or remote_preset.get('sudo', None)

        host = host or 'localhost'
        user = user or 'root'

        self.remote = Remote(host, user, password, public_key, sudo=sudo, debug=debug)

    def uninstall(self,
                  rpm: str = None,
                  ubrokerd_path: str = '/opt/universal/ubroker/ubrokerd',
                  paths: list = []):
        '''
        Uninstalls UBroker.

        If an RPM is provided, it will be uninstalled with `rpm -e`.
        Otherwise, the Universal Broker will be stopped and the paths
        provided will be removed.
        '''
        print('🟪[Uninstalling UBroker]')
        print.indent(close='\n')

        # Preset precedence
        conf = self.preset_data.get('uninstall', {})
        rpm = rpm or conf.get('rpm', None)
        ubrokerd_path = ubrokerd_path or conf.get('ubrokerd_path', None)
        paths = paths or conf.get('paths', None)

        if rpm:
            with print(f'⏹Uninstalling with RPM - {rpm}'):
                try:
                    self.remote.run(f'rpm -e {rpm}')
                except RuntimeError as e:
                    if 'is not installed' in str(e):
                        print('🟩Package not installed')
        elif ubrokerd_path:
            self.remote.run(f'{ubrokerd_path} stop')

        if paths:
            with print('⏹Cleaning up paths'):
                for path in paths:
                    with print(f'- {path}'):
                        self.remote.run(f'rm -r {path}')
        print.dedent()

    def install(self,
                link: str = None,
                release: str = None,
                match: str = None,
                options: dict = {}):
        '''Downloads and installs UBroker using the provided options'''
        print('🟪[Installing UBroker]')
        print.indent()

        working_dir = self.download_and_extract(link=link, release=release, match=match)

        options = options or self.preset_data.get('install', {}).get('options', {})
        with print('⏹Running installer script', indent='  | '):
            binary_options = ' '.join(options)
            sudo = 'sudo' if self.remote.sudo else ''
            out, err = self.remote.run(f'cd {working_dir} && {sudo} ./unvinst {binary_options}', sudo=False)
            print(out)
            if 'error' in err:
                raise RuntimeError(err)
        print.dedent()

    def download_and_extract(self, link: str = None, release: str = None, match: str = None):
        '''Downloads and extracts the Universal Broker installer tar to a temporary directory'''

        # Preset precedence
        preset = self.preset_data.get('install', {})
        link = link or preset.get('link', None)
        release = release or preset.get('release', None)
        match = match or preset.get('match', None)

        if not link and not release:
            raise ValueError('No link or release provided')

        # Get the link to download
        if not link:
            link = builds.get_build_link(release, match)

        # Setup working directory
        package = link.split('/')[-1]
        working_dir = '/tmp/uinstall'
        with print('⏹Setting up working directory', working_dir):
            self.remote.run(f'rm -r {working_dir}')
            self.remote.run_safe(f'mkdir -p {working_dir}')

        # Download and extract
        output_file = f'{working_dir}/uinstall-{package}'
        with print('⏹Downloading', link):
            _, err = self.remote.run(f'curl -o {output_file} {link}')
            if 'error' in err:
                raise RuntimeError(err)

        with print('⏹Extracting', output_file):
            self.remote.run_safe(f'tar -xzf {output_file} -C {working_dir}')

        return working_dir

    def load_preset(self, path: Path):
        if not path.exists() or not path.is_file():
            print("🟥Preset file not found")
            exit(2)

        try:
            self.preset_data = yaml.safe_load(path.read_text())
        except Exception as e:
            print(f"🟥Error loading preset: {e}")
            exit(3)

        print(f'🟩Loaded preset: "{path.resolve()}"')


def main():
    parser = ArgumentParser()
    parser.add_argument('preset', type=Path)
    parser.add_argument('-i', '--host', type=str, default=os.environ.get('SBUTILS_HOST'))
    parser.add_argument('-u', '--user', type=str, default=os.environ.get('SBUTILS_USER'))
    parser.add_argument('-p', '--password', type=str, default=os.environ.get('SBUTILS_PASSWORD'))
    parser.add_argument('-P', '--public-key', type=str, default=os.environ.get('SBUTILS_PUBLIC_KEY'))
    parser.add_argument('-s', '--no-sudo', action='store_false', dest='sudo')
    parser.add_argument('--release', type=str)
    parser.add_argument('--link', type=str)
    parser.add_argument('--match', type=str)
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args()
    if args.release and not re.match(r'^\d+\.\d+\.\d+\.\d+$', args.release):
        parser.error(f'Invalid release format: {args.release}, expected X.X.X.X')

    uinstall = UInstall(preset=args.preset,
                        host=args.host,
                        user=args.user,
                        password=args.password,
                        public_key=args.public_key,
                        sudo=args.sudo,
                        debug=args.debug)
    uinstall.uninstall()
    uinstall.install(link=args.link, release=args.release, match=args.match)


if __name__ == '__main__':
    main()



import os
import textwrap
from getpass import getpass

if os.name == 'nt':
    os.system('color')

_indent_stack = []
_close_stack = []
_print = print
_input = input


def get_color_escape(r, g, b, background=False):
    return '\033[{};2;{};{};{}m'.format(48 if background else 38, r, g, b)


RESET = '\033[0m'
RED = get_color_escape(239, 51, 77)
GREEN = get_color_escape(6, 180, 120)
BLUE = get_color_escape(17, 138, 178)
YELLOW = get_color_escape(255, 209, 102)
MAGENTA = get_color_escape(170, 74, 178)
CYAN = get_color_escape(108, 200, 230)
ORANGE = get_color_escape(255, 127, 17)
WHITE = get_color_escape(255, 255, 255)

colors = {
    '🟥': RED,
    '🟧': ORANGE,
    '🟨': YELLOW,
    '🟩': GREEN,
    '⏹': CYAN,
    '🟦': BLUE,
    '🟪': MAGENTA,
    '⬜': WHITE,
    '⬛': RESET
}


class IndentManager:

    def indent(self, indent: str = '    ', close=''):
        _indent_stack.append(indent)
        _close_stack.append(close)

    def dedent(self):
        _indent_stack.pop()
        _close_stack.pop()

    def __enter__(self):
        _indent_stack.append(self._next_indent)
        _close_stack.append(self._next_close)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _indent_stack.pop()
        _print(_close_stack.pop(), end='')

    def _color(self, text):
        colored = False
        for color, escape in colors.items():
            if color in text:
                text = text.replace(color, escape)
                colored = True

        if colored:
            text += RESET
        return text

    def __call__(self, *args, indent: str = '    ', close='', **kwargs):
        text = ' '.join(map(str, args))
        text = self._color(text)
        text = textwrap.indent(text, ''.join(_indent_stack))
        _print(text, **kwargs)
        self._next_indent = indent
        self._next_close = close
        return self

    def input(self, prompt: str = '', password: bool = False):
        prompt = textwrap.indent(prompt, ''.join(_indent_stack))
        prompt = self._color(prompt)
        if password:
            return getpass(prompt)
        return _input(prompt)


_inter = IndentManager()
print = _inter
input = print.input



import re
import requests

from .printing import print, input


def get_build_link(release: str, substring: str) -> str:
    '''
    Get the build link for the given release.
    '''
    release_short = release.replace('.', '')[:3]
    page = f'http://build.stone.branch/rel/agent/release/RB_WA_{release_short}/current/'
    links = listings(page, substring)
    if not links:
        raise ValueError(f'No builds found for release {release} and matching {substring}')

    if len(links) == 1:
        return page + links[0]

    print('Available builds:\n')
    for i, link in enumerate(links):
        print(f'{i+1}'.ljust(3), ':', link)
    print()
    choice = input('ACTION REQUIRED >> Enter the number of the build to download: ')
    print()
    return page + links[int(choice)-1]


def listings(url, substring):
    response = requests.get(url).text
    response = response.replace('\\n', '\n')
    links = re.findall('href="(.+?)"', response)
    result = []
    for link in links:
        if substring in link and not link.endswith('.asc'):
            result.append(link)
    return result



import re
import os
import yaml
import difflib
import fnmatch
import itertools
from argparse import ArgumentParser
from pathlib import Path

from .printing import print
from .remote import Remote


class UConf:

    def __init__(self,
                 preset: Path = None,
                 host: str = None,
                 user: str = None,
                 password: str = None,
                 public_key: str = None,
                 sudo: bool = False,
                 debug: bool = False,
                 directory: str = None):

        self.preset_data = {}
        if preset:
            with print('🟪[Loading preset]', close='\n'):
                self.load_preset(preset)

        self.debug = debug
        self.directory = directory

        if self.preset_data.get('remote', None):
            remote_preset = self.preset_data['remote']
            host = host or remote_preset.get('host', None)
            user = user or remote_preset.get('user', None)
            public_key = public_key or remote_preset.get('public_key', None)
            sudo = sudo or remote_preset.get('sudo', None)

        host = host or 'localhost'
        user = user or 'root'

        self.remote = Remote(host, user, password, public_key, sudo=sudo, debug=debug)

    def list_options(self, config: str = None, directory: str = None, comp: bool = False):
        '''
        List the available configuration options.
        '''
        path = self.get_config_path(config, directory, comp)
        with print(f'🟪{path}'):
            content = self.remote.get(path)

            # Find all options and defaults
            options = re.findall(r'#\s+Option:\s+(.*)', content, re.MULTILINE)
            defaults = re.findall(r'#\s+[Dd]efault:\s+(.*)', content, re.MULTILINE)

            longest = max([len(option.strip()) for option in options])
            lines = []
            for option, default in zip(options, defaults):
                lines.append(f'{option.strip()} {"." * (longest - len(option) + 4)} {default.strip()}')

            lines = sorted(lines)
            # split each option by _ and group them into separate lists by the first part
            groups = [list(g) for k, g in itertools.groupby(lines, lambda x: x.split('_')[0])]
            # sort groups by number of options in each group, then alphabetically
            groups.sort(key=lambda x: (len(x), x[0][0]))

            option = "[Option Name]"
            print(f'🟩{option} {" " * (longest - len(option) + 4)} [Default Value]')
            for i, group in enumerate(groups):
                for option in group:
                    option = option.strip()
                    option = f'🟦{option}' if i % 2 == 0 else f'⏹{option}'
                    print(option)

    def get_option_info(self, config: str, option: str, directory: str = None, comp: bool = False):
        '''
        Get the description of a specific configuration option.
        '''
        config_path = self.get_config_path(config, directory, comp)
        content = self.remote.get(config_path)
        lines = content.splitlines()

        found = False
        for line in lines:
            if found and line.startswith('# Option') or line.startswith('#####'):
                break
            if re.match(r'#\s+Option:\s+{}'.format(option), line):
                found = True
            if found:
                print(f'⏹{line[1:].rstrip()}')
        else:
            print(f'🟨Option {option} not found in {config_path}')

    def update(self, config: str, directory: str = None, comp: bool = False, **options: str):
        '''
        Set configuration options in a configuration file.
        '''
        config_path = self.get_config_path(config, directory, comp)

        with print(f'🟪{config_path}'):
            content = self.remote.get(config_path)
            current_options = self.parse(content)
            new_content = content.splitlines()

            for option, value in options.items():
                if value is not None:
                    value = '{0}    {1}'.format(" " * max(24 - len(option), 0), value)
                matches = fnmatch.filter(current_options.keys(), option)
                if matches:
                    for match in matches:
                        spanned_lines = current_options[match]['lines']
                        for i in spanned_lines:
                            new_content[i] = None
                        if value is not None:
                            new_content[spanned_lines[0]] = '{0}{1}\n'.format(match, value)
                elif value is not None:
                    new_content.append('{0}{1}\n'.format(option, value))
            new_content = '\n'.join([i for i in new_content if i is not None])
            self.remote.put(config_path, new_content)
            self.diff(content, new_content)

    def diff(self, original, updated):
        # remove comments
        orig = [line for line in original.splitlines() if not line.startswith('#')]
        new = [line for line in updated.splitlines() if not line.startswith('#')]

        # get the diff

        result = difflib.Differ().compare(orig, new)

        changed = False
        for line in result:
            if line.startswith('?'):
                line = line.replace('?', '🟨?', 1)
                changed = True
            if line.startswith('-'):
                line = line.replace('-', '🟥-', 1)
                changed = True
            if line.startswith('+'):
                line = line.replace('+', '🟩+', 1)
                changed = True
            print(line)

        if not changed:
            print('⏹No changes were made')

    def parse(self, content: str) -> dict:
        compacted = []
        multiline = False
        buffer = " "
        line = ""
        for line_num, line in enumerate(content.splitlines()):
            if line.startswith('#'):
                continue

            line = line.strip()
            if len(line) == 0:
                continue
            last_char = line[-1]
            if last_char == '-' or last_char == '+':
                line = line[0:-1]

            if multiline:
                compacted[-1]['entry'] = (compacted[-1]['entry'] + buffer + line)
                compacted[-1]['lines'].append(line_num)
            else:
                compacted.append({'entry': line, 'lines': [line_num]})

            if last_char == '+':
                multiline = True
                buffer = ""
            elif last_char == '-':
                multiline = True
                buffer = " "
            else:
                multiline = False

        result = {}
        for item in compacted:
            pair = item['entry'].split(maxsplit=1)
            if len(pair) != 2:
                result[pair[0]] = {'value': None, 'lines': item['lines']}
            else:
                result[pair[0]] = {'value': pair[1], 'lines': item['lines']}
        return result

    def get_config_path(self, name: str, directory: str = None, comp: bool = False):
        directory = directory or self.directory or '/etc/universal/'
        comp = comp or self.preset_data.get('comp', False)

        if not comp and not name.endswith('.conf'):
            name += '.conf'

        return directory.rstrip('/') + '/' + name

    def load_preset(self, path: Path):
        if not path.exists() or not path.is_file():
            print("🟥Preset file not found")
            exit(2)

        try:
            self.preset_data = yaml.safe_load(path.read_text())
        except Exception as e:
            print(f"🟥Error loading preset: {e}")
            exit(3)

        print(f'🟩Loaded preset: "{path.resolve()}"')


def main():
    parser = ArgumentParser()

    parser.add_argument('config_or_preset', type=str)
    parser.add_argument('options', nargs='*', )

    # Remote
    parser.add_argument('-i', '--host', type=str, default=os.environ.get('SBUTILS_HOST'))
    parser.add_argument('-u', '--user', type=str, default=os.environ.get('SBUTILS_USER'))
    parser.add_argument('-p', '--password', type=str, default=os.environ.get('SBUTILS_PASSWORD'))
    parser.add_argument('-P', '--public-key', type=str, default=os.environ.get('SBUTILS_PUBLIC_KEY'))
    parser.add_argument('-s', '--no-sudo', action='store_false', dest='sudo')
    parser.add_argument('--debug', action='store_true')
    # Function
    parser.add_argument('-d', '--directory', type=str)
    parser.add_argument('-m', '--more', type=str)
    parser.add_argument('-c', '--comp', action='store_true')
    parser.add_argument('-l', '--list', action='store_true')
    parser.add_argument('-B', '--backup', action='store_true')
    parser.add_argument('-R', '--restore', action='store_true')

    args = parser.parse_args()

    if args.config_or_preset.endswith('.yml'):
        preset = args.config_or_preset
        config = None
    else:
        config = args.config_or_preset
        preset = None

    uconf = UConf(preset=preset,
                  host=args.host,
                  user=args.user,
                  password=args.password,
                  public_key=args.public_key,
                  sudo=args.sudo,
                  debug=args.debug,
                  directory=args.directory)
    if preset:
        # uconf.load_preset(preset)
        pass
    elif args.list:
        uconf.list_options(config=config, directory=args.directory, comp=args.comp)
    elif args.more:
        uconf.get_option_info(config=config, option=args.more, directory=args.directory, comp=args.comp)
    elif args.backup:
        pass
    elif args.restore:
        pass
    else:
        options = {}
        for option in args.options:
            if '=' not in option:
                options[option] = None
                continue
            key, value = option.split('=', 1)
            options[key] = value
        uconf.update(config=config, directory=args.directory, comp=args.comp, **options)


if __name__ == '__main__':
    main()



import paramiko
import subprocess
import socket
import struct
from elevate import elevate

from .printing import print, input


class Remote:

    def __init__(self,
                 host: str,
                 username: str,
                 password: str = None,
                 public_key: str = None,
                 port: int = 22,
                 sudo: bool = False,
                 debug: bool = False):
        self.host = host
        self.username = username
        self.public_key = public_key
        self.port = port
        self.sudo = sudo
        self.debug = debug
        self.is_local = False

        if self.is_loopback(host):
            self.is_local = True
            if self.sudo:
                elevate(graphical=False)
                self.sudo = False
            return

        with print(f'🟪[Connecting to {host} as {username}]', close='\n'):
            self.password = password or input(f'🟨Enter password for {username}@{host}: ', password=True)
            self.client = self.create_client()

    def create_client(self):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(self.host,
                    username=self.username,
                    password=self.password,
                    key_filename=self.public_key,
                    allow_agent=True,
                    port=self.port)
        print('🟩Connected')
        return ssh

    def run(self, command, sudo=...):
        if sudo or self.sudo and sudo is ...:
            command = f'sudo {command}'
        if self.debug:
            print(f'🟩$ {command}')
            print.indent()

        if self.is_local:
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()
            stdout = stdout.decode('utf-8').strip()
            stderr = stderr.decode('utf-8').strip()
        else:
            stdin, stdout, stderr = self.client.exec_command(command)
            stdout = stdout.read().decode('utf-8').strip()
            stderr = stderr.read().decode('utf-8').strip()

        if self.debug:
            if stdout:
                with print('STDOUT', indent='  | '):
                    print(stdout)
            if stderr:
                with print('STDERR', indent='  | '):
                    print(stderr)
            print.dedent()
        return stdout, stderr

    def run_safe(self, command):
        stdout, stderr = self.run(command)
        if stderr:
            raise RuntimeError(stderr)
        return stdout

    def get(self, path) -> str:
        if self.is_local:
            with open(path) as f:
                return f.read()
        else:
            with self.client.open_sftp().file(path, 'r') as f:
                return f.read().decode('utf-8', errors='replace')

    def put(self, path, content):
        if self.is_local:
            with open(path, 'w') as f:
                f.write(content)
        else:
            with self.client.open_sftp().file(path, 'w') as f:
                f.write(content)

    def is_loopback(self, host: str) -> bool:
        loopback_checker = {
            socket.AF_INET: lambda x: struct.unpack('!I', socket.inet_aton(x))[0] >> (32-8) == 127,
            socket.AF_INET6: lambda x: x == '::1'
        }
        for family in (socket.AF_INET, socket.AF_INET6):
            try:
                r = socket.getaddrinfo(host, None, family, socket.SOCK_STREAM)
            except socket.gaierror:
                return False
            for family, _, _, _, sockaddr in r:
                if not loopback_checker[family](sockaddr[0]):
                    return False
        return True



from setuptools import setup, find_packages


def main():
    requires = [
        'requests',
        'paramiko',
        'pyyaml',
        'prodict',
    ]

    setup(
        name='sbutils-remote',
        version='0.0.2',
        packages=find_packages(exclude=[]),
        include_package_data=True,
        install_requires=requires,
        author='Stonebranch',
        description='SBUtils but for remote servers',
        entry_points={
            'console_scripts': [
                'rconf=src.uconf:main',
                'rinstall=src.uinstall:main',
            ]
        },
        python_requires='>=2.7',
        classifiers=[
            'Operating System :: POSIX :: Linux',
            'Operating System :: Microsoft :: Windows',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3'
        ],
        long_description='',
        long_description_content_type="text/x-rst"
    )


if __name__ == '__main__':
    main()
